import { SemaphoreInvalidPermitsException } from "./exceptions";
import { ConcurrencyExceedTimeoutException, ConcurrencyInvalidTimeoutException } from "../exceptions";
import { ConcurrencyInterruptedException } from "../exceptions";
var Semaphore = (function () {
    function Semaphore(permits) {
        if (permits === void 0) { permits = 1; }
        this.permits = permits;
        this.queue = [];
        if (permits < 0) {
            throw new SemaphoreInvalidPermitsException("Permits must be >= to 0. '".concat(permits, "' given."));
        }
    }
    Object.defineProperty(Semaphore.prototype, "permitsAvailable", {
        get: function () {
            return this.permits;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Semaphore.prototype, "permitsRequired", {
        get: function () {
            return this.queue.reduce(function (total, _a) {
                var resolvers = _a.resolvers;
                return total + resolvers.length;
            }, 0);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Semaphore.prototype, "queueLength", {
        get: function () {
            return this.queue.length;
        },
        enumerable: false,
        configurable: true
    });
    Semaphore.prototype.acquire = function (permits) {
        if (permits === void 0) { permits = 1; }
        if (permits < 0) {
            return Promise.reject(new SemaphoreInvalidPermitsException("Cannot acquire negative permits"));
        }
        while (this.permits > 0 && permits) {
            --this.permits;
            --permits;
        }
        if (permits === 0) {
            return Promise.resolve();
        }
        return this.acquireLock(permits);
    };
    Semaphore.prototype.tryAcquire = function (timeout, permits) {
        var _this = this;
        if (permits === void 0) { permits = 1; }
        if (timeout < 0) {
            return Promise.reject(new ConcurrencyInvalidTimeoutException("Cannot acquire with a negative timeout"));
        }
        if (permits < 0) {
            return Promise.reject(new SemaphoreInvalidPermitsException("Cannot acquire negative permits"));
        }
        if (permits <= this.permits) {
            this.permits -= permits;
            return Promise.resolve();
        }
        var permitsRemaining = permits - this.permits;
        var permitsBck = this.permits;
        this.permits = 0;
        var timeoutId = null;
        return this.acquireLock(permitsRemaining, function (item) {
            var reject = item.reject, resolvers = item.resolvers;
            timeoutId = setTimeout(function () {
                var index = _this.queue.findIndex(function (i) { return i === item; });
                if (index >= 0) {
                    _this.queue.splice(index, 1);
                }
                var permitsTaken = permitsBck + (permitsRemaining - resolvers.length);
                var permitsToRelease = Math.min(permitsTaken, _this.permitsRequired);
                _this.permits += permitsTaken - permitsToRelease;
                if (permitsToRelease > 0) {
                    _this.release(permitsToRelease);
                }
                reject(new ConcurrencyExceedTimeoutException("Timeout of ".concat(timeout, "ms exceed when acquiring.")));
            }, timeout);
        }).finally(function () {
            if (timeoutId) {
                clearTimeout(timeoutId);
            }
        });
    };
    Semaphore.prototype.release = function (permits) {
        if (permits === void 0) { permits = 1; }
        if (permits < 0) {
            throw new SemaphoreInvalidPermitsException("Cannot release negative permits");
        }
        if (permits === 0) {
            return;
        }
        if (this.queue.length) {
            var resolvers = this.queue[0].resolvers;
            resolvers.splice(0, 1)[0]();
            if (!resolvers.length) {
                this.queue.splice(0, 1);
            }
        }
        else {
            ++this.permits;
        }
        this.release(permits - 1);
    };
    Semaphore.prototype.releaseAll = function (permits) {
        if (permits === void 0) { permits = 0; }
        if (permits < 0) {
            throw new SemaphoreInvalidPermitsException("Cannot set negative permits");
        }
        for (var _i = 0, _a = this.queue.splice(0).flatMap(function (_a) {
            var resolvers = _a.resolvers;
            return resolvers;
        }); _i < _a.length; _i++) {
            var resolver = _a[_i];
            resolver();
        }
        this.permits = permits;
    };
    Semaphore.prototype.interrupt = function (reason, permits) {
        if (permits === void 0) { permits = 0; }
        if (permits < 0) {
            throw new SemaphoreInvalidPermitsException("Cannot set negative permits");
        }
        var exception = new ConcurrencyInterruptedException(reason, "The semaphore has been interrupted.");
        for (var _i = 0, _a = this.queue.splice(0); _i < _a.length; _i++) {
            var reject = _a[_i].reject;
            reject(exception);
        }
        this.permits = permits;
    };
    Semaphore.prototype.acquireLock = function (permits, afterInit) {
        var _this = this;
        return new Promise(function (resolve, reject) {
            var resolvers = [];
            void Promise.all(Array(permits)
                .fill(null)
                .map(function () { return new Promise(function (resolve) { return resolvers.push(resolve); }); })).then(function () { return resolve(); });
            var item = { reject: reject, resolvers: resolvers };
            _this.queue.push(item);
            if (afterInit) {
                afterInit(item);
            }
        });
    };
    return Semaphore;
}());
export { Semaphore };
//# sourceMappingURL=semaphore.js.map