import { Synchronizer } from "../synchronizer.interface";
/**
 * A semaphore to manage "concurrency in Javascript" (as mentioned in the mozilla documentation).
 *
 * The queue is a simple FIFO (First In, First Out).
 *
 * @see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise#promise_concurrency
 */
export declare class Semaphore implements Synchronizer {
    private permits;
    /**
     * Queue of the "threads"
     */
    private readonly queue;
    /**
     * @returns the current number of available permits
     */
    get permitsAvailable(): number;
    /**
     * @returns the minimal number of permits needed to release all "Threads"
     */
    get permitsRequired(): number;
    /**
     * @returns the number of waiting "Threads" on an acquire or tryAcquire (!= number of permits awaited)
     */
    get queueLength(): number;
    /**
     * Creates a Semaphore
     *
     * @param permits the initials permits
     * @throws {SemaphoreInvalidPermitsException} when the `permits` value is invalid
     */
    constructor(permits?: number);
    /**
     * Acquires an amount of permits:
     * Wait until the given number of permits is available.
     *
     * @param permits The number of permits to acquire (0 acquire immediately)
     * @throws {SemaphoreInvalidPermitsException} when the `permits` value is invalid
     * @returns a promise when the permits have been acquired
     */
    acquire(permits?: number): Promise<void>;
    /**
     * Acquires a given amount of permits:
     * Wait until the given number of permits is available.
     *
     * Throws an error if the given time exceeds
     * and re-establish the state as if the method was not called.
     *
     * @param timeout maximum time (in ms) to acquire the permits
     * @param permits The number of permits to acquire (0 acquire immediately)
     * @throws {ConcurrencyExceedTimeoutException} when the `timeout` value is invalid
     * @throws {SemaphoreInvalidPermitsException} when the `permits` value is invalid
     * @returns a promise when the permits have been acquired
     */
    tryAcquire(timeout: number, permits?: number): Promise<void>;
    /**
     * Release a given amount of permits.
     *
     * @param permits The number of permits to release (0 does nothing)
     * @throws {SemaphoreInvalidPermitsException} when the `permits` value is invalid
     */
    release(permits?: number): void;
    /**
     * Releases all awaiting "Threads".
     *
     * @param permits The permits to set to this semaphore once everything has been released
     * @throws {SemaphoreInvalidPermitsException} when the `permits` value is invalid
     * @see {interrupt} to throw an exception
     */
    releaseAll(permits?: number): void;
    /**
     * Interrupts all awaiting "threads" with an [exception]{@link ConcurrencyInterruptedException}.
     *
     * @param reason The reason why this semaphore is being interrupted
     * @param permits The permits to set to this semaphore once everything has been interrupted
     * @throws {SemaphoreInvalidPermitsException} when the `permits` value is invalid
     * @see {releaseAll} to not throw an exception
     */
    interrupt(reason: unknown, permits?: number): void;
    /**
     * @param permits the permits to obtain
     * @param afterInit after the blocking promise resolver has been set but still in the promise body
     * @returns the blocking promise
     */
    private acquireLock;
}
